package pod

import (
	containerLib "github.com/ankitrgadiya/operatorlib/pkg/container"
	"github.com/ankitrgadiya/operatorlib/pkg/meta"

	"github.com/pkg/errors"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
)

// GenerateConfigMap generates ConfigMap object as per the `Conf` struct passed
func GeneratePod(c Conf) (obj *corev1.PodTemplateSpec, err error) {
	var om *metav1.ObjectMeta
	var container corev1.Container
	var containers []corev1.Container
	var initContainer corev1.Container
	var initContainers []corev1.Container
	var volumes []corev1.Volume
	var securityContext *corev1.PodSecurityContext
	var restartPolicy corev1.RestartPolicy

	om, err = meta.GenerateObjectMeta(meta.Conf{
		Instance:           c.Instance,
		Name:               c.Name,
		Namespace:          c.Namespace,
		GenLabelsFunc:      c.GenLabelsFunc,
		GenAnnotationsFunc: c.GenAnnotationsFunc,
		AppendLabels:       c.AppendLabels,
	})

	if err != nil {
		return nil, errors.Wrap(err, "failed to generate objectmeta")
	}

	if c.Containers != nil {
		for _, genContainer := range c.Containers {
			cont, err := containerLib.GenerateContainer(containerLib.Conf{
				Instance:                       c.Instance,
				Namespace:                      c.Namespace,
				Name:                           genContainer.Name,
				Image:                          genContainer.Image,
				ImagePullPolicy:                genContainer.ImagePullPolicy,
				GenContainerPortsFunc:          genContainer.GenContainerPortsFunc,
				GenContainerCmdFunc:            genContainer.GenContainerCmdFunc,
				GenContainerArgsFunc:           genContainer.GenContainerArgsFunc,
				GenContainerResourcesFunc:      genContainer.GenContainerResourcesFunc,
				GenContainerVolumeMountsFunc:   genContainer.GenContainerVolumeMountsFunc,
				GenContainerReadinessProbeFunc: genContainer.GenContainerReadinessProbeFunc,
				GenContainerEnvVarFunc:         genContainer.GenContainerEnvVarFunc,
			})
			if err != nil {
				return nil, errors.Wrap(err, "failed to generate pod template spec")
			}
			container = *cont
			containers = append(containers, container)
		}
	}

	if c.InitContainers != nil {
		for _, genContainer := range c.InitContainers {
			initCont, err := containerLib.GenerateContainer(containerLib.Conf{
				Instance:                       c.Instance,
				Namespace:                      c.Namespace,
				Name:                           genContainer.Name,
				Image:                          genContainer.Image,
				ImagePullPolicy:                genContainer.ImagePullPolicy,
				GenContainerPortsFunc:          genContainer.GenContainerPortsFunc,
				GenContainerCmdFunc:            genContainer.GenContainerCmdFunc,
				GenContainerArgsFunc:           genContainer.GenContainerArgsFunc,
				GenContainerResourcesFunc:      genContainer.GenContainerResourcesFunc,
				GenContainerVolumeMountsFunc:   genContainer.GenContainerVolumeMountsFunc,
				GenContainerReadinessProbeFunc: genContainer.GenContainerReadinessProbeFunc,
				GenContainerEnvVarFunc:         genContainer.GenContainerEnvVarFunc,
			})
			if err != nil {
				return nil, errors.Wrap(err, "failed to generate pod template spec")
			}
			initContainer = *initCont
			initContainers = append(initContainers, initContainer)
		}
	}

	if c.GenVolumesFunc != nil {
		volumes, err = c.GenVolumesFunc(c.Instance)
		if err != nil {
			return nil, errors.Wrap(err, "failed to generate volumes for pod template spec")
		}
	}

	if c.GenSecurityContextFunc != nil {
		securityContext, err = c.GenSecurityContextFunc(c.Instance)
		if err != nil {
			return nil, errors.Wrap(err, "failed to generate securityContext pod template spec")
		}
	}

	if c.GenRestartPolicyFunc != nil {
		restartPolicy, err = c.GenRestartPolicyFunc(c.Instance)
		if err != nil {
			return nil, errors.Wrap(err, "failed to generate securityContext pod template spec")
		}
	}

	obj = &corev1.PodTemplateSpec{
		ObjectMeta: *om,
		Spec: corev1.PodSpec{
			InitContainers:  initContainers,
			Containers:      containers,
			Volumes:         volumes,
			SecurityContext: securityContext,
			RestartPolicy:   restartPolicy,
		},
	}

	return obj, err
}
